import decimal

from collections import namedtuple

__all__ = ("Currency", "set_default_currency", "Amount", "NO_CURRENCY")


class Currency(namedtuple("Currency", "code number name symbol precision")):
    def __new__(cls, code, number, name="", symbol=u"", precision=2):
        return super(Currency, cls).__new__(cls, code, number, name, symbol, precision)

    def __eq__(self, other):
        if isinstance(other, Currency):
            return self.code == other.code
        if isinstance(other, str):
            return self.code == other
        return NotImplemented

    def __repr__(self):
        return "Currency({!r}, {!r})".format(self.code, self.number)

    def __str__(self):
        if self.name:
            return "{} - {}".format(self.name, self.code)
        else:
            return self.code


CURRENCY = {}
# XXX is the currency code for no currency in ISO 4217
CURRENCY["XXX"] = DEFAULT_CURRENCY = NO_CURRENCY = Currency(
    "XXX", 999, "No Currency", "", 4
)


def set_default_currency(code="XXX"):
    """
    Set the default currency (the one applied if no currency is specified)
    :param code:
    :return:
    """
    global DEFAULT_CURRENCY
    DEFAULT_CURRENCY = CURRENCY[code]


class Amount(tuple):
    """
    A monetary amount and the associated currency.
    """

    def __new__(cls, value=None, currency=None):
        if isinstance(value, (tuple, list)):
            if len(value) == 1:
                value = value[0]
            elif len(value) == 2:
                value, currency = value
            else:
                raise ValueError(
                    "expected tuple with a length of 1 or 2 got: {}".format(len(value))
                )

        if not isinstance(value, decimal.Decimal):
            try:
                value = decimal.Decimal(value)
            except decimal.InvalidOperation:
                raise ValueError(
                    "value is not a valid numerical amount: {}".format(value)
                )

        if currency is None:
            currency = DEFAULT_CURRENCY
        elif isinstance(currency, str):
            currency = CURRENCY[currency.upper()]
        elif not isinstance(currency, Currency):
            raise TypeError(
                "unable to convert value into a valid currency: {}".format(currency)
            )

        return tuple.__new__(cls, (value, currency))

    # Type conversions

    def __float__(self):
        return float(self.value)

    def __int__(self):
        return int(self.value)

    def __str__(self):
        if self.currency == NO_CURRENCY:
            return self.format("{value}")
        else:
            return self.format("{value} {currency.code}")

    def __repr__(self):
        return self.format("<Amount: {value}, {currency!r}>")

    # Math operations

    def __round__(self, n=None):
        """
        Round the value to the specified.
        """
        return self.value.__round__(n or self.currency.precision)

    def __pos__(self):
        return Amount(self.value, self.currency)

    def __neg__(self):
        return Amount(-self.value, self.currency)

    def __add__(self, other):
        other_value, currency = self._pre_calculate(other)
        return Amount(self.value + other_value, currency)

    def __sub__(self, other):
        other_value, currency = self._pre_calculate(other)
        return Amount(self.value - other_value, currency)

    def __mul__(self, other):
        if isinstance(other, Amount):
            raise TypeError("monetary amounts can not be multiplied.")
        else:
            return Amount(self.value * decimal.Decimal(other), self.currency)

    def __truediv__(self, other):
        other_value, currency = self._pre_calculate(other)
        # If the other instance is a currency we return just a value.
        if isinstance(other, Amount):
            return self.value / other_value
        else:
            return Amount(self.value / other_value, currency)

    __radd__ = __add__
    __rsub__ = __sub__
    __rmul__ = __mul__
    __div__ = __rtruediv__ = __truediv__

    # Comparison operators

    def __eq__(self, other):
        if isinstance(other, Amount):
            return self.value == other.value and self.currency == other.currency
        return NotImplemented

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        other_value, _ = self._pre_calculate(other)
        return self.value < other_value

    def __gt__(self, other):
        other_value, _ = self._pre_calculate(other)
        return self.value > other_value

    def __le__(self, other):
        return self < other or self == other

    def __ge__(self, other):
        return self > other or self == other

    def _pre_calculate(self, other):
        """
        Handle any pre-calculation checks
        """
        if isinstance(other, Amount):
            other_value = other.value
            other_currency = other.currency
        else:
            other_value = decimal.Decimal(other)
            other_currency = NO_CURRENCY

        if self.currency == NO_CURRENCY:
            currency = other_currency
        elif other_currency == NO_CURRENCY:
            currency = self.currency
        elif self.currency != other.currency:
            raise ValueError(
                "Cannot perform calculation on amounts of different currencies."
            )
        else:
            currency = self.currency

        return other_value, currency

    @property
    def value(self):
        return self[0]

    @property
    def currency(self):
        return self[1]

    @property
    def is_naive(self):
        """
        Boolean indicating if this field is assigned `NO_CURRENCY`.
        """
        return self.currency == NO_CURRENCY

    def format(self, format_string):
        """
        Format an amount.

        Internally this uses the python builtin string format method. The following fields are available for formatting:

        - *value*; value of the amount formatted to the precision defined by the currency
        - *value_raw*; the raw value used internally (a Decimal)
        - *currency*; the currency
          - *currency.symbol*; the symbol defined by the currency
          - *currency.code*; the code defined by the currency
          - *currency.number*; the number defined by the currency (currencies are assigned a number along with a code)
          - *currency.name*; the name defined by the currency

        Usage::

            > Amount(10, 'USD').format("{symbol}{value} {code}")
            $10.00 USD

        :param format_string:
        :return: formatted string

        """
        assert isinstance(format_string, str)
        return format_string.format(
            value=("{{0:0.{:d}f}}".format(self.currency.precision)).format(self.value),
            value_raw=self.value,
            currency=self.currency,
        )

    def assign_currency(self, currency):
        """
        Assign a currency if this amount currently has no currency assigned.

        As an ``Amount`` is immutable this will return a new ``Amount`` instance.

        :param currency: Currency to apply.
        :return:

        """
        if not self.is_naive:
            raise ValueError("This amount is already assigned a currency.")
        return Amount(self.value, CURRENCY[currency])


# Currencies based off the latest ISO 4217 list of currencies
# Updated: 2014-08-29
CURRENCY["AED"] = Currency("AED", 784, "UAE Dirham", u"", 2)
CURRENCY["AFN"] = Currency("AFN", 971, "Afghani", u"؋", 2)
CURRENCY["ALL"] = Currency("ALL", 8, "Lek", u"", 2)
CURRENCY["AMD"] = Currency("AMD", 51, "Armenian Dram", u"", 2)
CURRENCY["ANG"] = Currency("ANG", 532, "Netherlands Antillean Guilder", u"ƒ", 2)
CURRENCY["AOA"] = Currency("AOA", 973, "Kwanza", u"", 2)
CURRENCY["ARS"] = Currency("ARS", 32, "Argentine Peso", u"$", 2)
CURRENCY["AUD"] = Currency("AUD", 36, "Australian Dollar", u"$", 2)
CURRENCY["AWG"] = Currency("AWG", 533, "Aruban Florin", u"ƒ", 2)
CURRENCY["AZN"] = Currency("AZN", 944, "Azerbaijanian Manat", u"ман", 2)
CURRENCY["BAM"] = Currency("BAM", 977, "Convertible Mark", u"KM", 2)
CURRENCY["BBD"] = Currency("BBD", 52, "Barbados Dollar", u"$", 2)
CURRENCY["BDT"] = Currency("BDT", 50, "Taka", u"", 2)
CURRENCY["BGN"] = Currency("BGN", 975, "Bulgarian Lev", u"", 2)
CURRENCY["BHD"] = Currency("BHD", 48, "Bahraini Dinar", u"", 3)
CURRENCY["BIF"] = Currency("BIF", 108, "Burundi Franc", u"", 0)
CURRENCY["BMD"] = Currency("BMD", 60, "Bermudian Dollar", u"", 2)
CURRENCY["BND"] = Currency("BND", 96, "Brunei Dollar", u"$", 2)
CURRENCY["BOB"] = Currency("BOB", 68, "Boliviano", u"", 2)
CURRENCY["BOV"] = Currency("BOV", 984, "Mvdol", u"", 2)
CURRENCY["BRL"] = Currency("BRL", 986, "Brazilian Real", u"", 2)
CURRENCY["BSD"] = Currency("BSD", 44, "Bahamian Dollar", u"", 2)
CURRENCY["BTN"] = Currency("BTN", 64, "Ngultrum", u"", 2)
CURRENCY["BWP"] = Currency("BWP", 72, "Pula", u"P", 2)
CURRENCY["BYR"] = Currency("BYR", 974, "Belarussian Ruble", u"p.", 0)
CURRENCY["BZD"] = Currency("BZD", 84, "Belize Dollar", u"BZ$", 2)
CURRENCY["CAD"] = Currency("CAD", 124, "Canadian Dollar", u"$", 2)
CURRENCY["CDF"] = Currency("CDF", 976, "Congolese Franc", u"", 2)
CURRENCY["CHE"] = Currency("CHE", 947, "WIR Euro", u"", 2)
CURRENCY["CHF"] = Currency("CHF", 756, "Swiss Franc", u"CHF", 2)
CURRENCY["CHW"] = Currency("CHW", 948, "WIR Franc", u"", 2)
CURRENCY["CLF"] = Currency("CLF", 990, "Unidad de Fomento", u"", 4)
CURRENCY["CLP"] = Currency("CLP", 152, "Chilean Peso", u"$", 0)
CURRENCY["CNY"] = Currency("CNY", 156, "Yuan Renminbi", u"¥", 2)
CURRENCY["COP"] = Currency("COP", 170, "Colombian Peso", u"$", 2)
CURRENCY["COU"] = Currency("COU", 970, "Unidad de Valor Real", u"", 2)
CURRENCY["CRC"] = Currency("CRC", 188, "Costa Rican Colon", u"₡", 2)
CURRENCY["CUC"] = Currency("CUC", 931, "Peso Convertible", u"", 2)
CURRENCY["CUP"] = Currency("CUP", 192, "Cuban Peso", u"₱", 2)
CURRENCY["CVE"] = Currency("CVE", 132, "Cabo Verde Escudo", u"", 2)
CURRENCY["CZK"] = Currency("CZK", 203, "Czech Koruna", u"Kč", 2)
CURRENCY["DJF"] = Currency("DJF", 262, "Djibouti Franc", u"", 0)
CURRENCY["DKK"] = Currency("DKK", 208, "Danish Krone", u"kr", 2)
CURRENCY["DOP"] = Currency("DOP", 214, "Dominican Peso", u"RD$", 2)
CURRENCY["DZD"] = Currency("DZD", 12, "Algerian Dinar", u"", 2)
CURRENCY["EGP"] = Currency("EGP", 818, "Egyptian Pound", u"£", 2)
CURRENCY["ERN"] = Currency("ERN", 232, "Nakfa", u"", 2)
CURRENCY["ETB"] = Currency("ETB", 230, "Ethiopian Birr", u"", 2)
CURRENCY["EUR"] = Currency("EUR", 978, "Euro", u"€", 2)
CURRENCY["FJD"] = Currency("FJD", 242, "Fiji Dollar", u"$", 2)
CURRENCY["FKP"] = Currency("FKP", 238, "Falkland Islands Pound", u"£", 2)
CURRENCY["GBP"] = Currency("GBP", 826, "Pound Sterling", u"£", 2)
CURRENCY["GEL"] = Currency("GEL", 981, "Lari", u"", 2)
CURRENCY["GHS"] = Currency("GHS", 936, "Ghana Cedi", u"¢", 2)
CURRENCY["GIP"] = Currency("GIP", 292, "Gibraltar Pound", u"£", 2)
CURRENCY["GMD"] = Currency("GMD", 270, "Dalasi", u"", 2)
CURRENCY["GNF"] = Currency("GNF", 324, "Guinea Franc", u"", 0)
CURRENCY["GTQ"] = Currency("GTQ", 320, "Quetzal", u"Q", 2)
CURRENCY["GYD"] = Currency("GYD", 328, "Guyana Dollar", u"$", 2)
CURRENCY["HKD"] = Currency("HKD", 344, "Hong Kong Dollar", u"$", 2)
CURRENCY["HNL"] = Currency("HNL", 340, "Lempira", u"L", 2)
CURRENCY["HRK"] = Currency("HRK", 191, "Croatian Kuna", u"kn", 2)
CURRENCY["HTG"] = Currency("HTG", 332, "Gourde", u"", 2)
CURRENCY["HUF"] = Currency("HUF", 348, "Forint", u"Ft", 2)
CURRENCY["IDR"] = Currency("IDR", 360, "Rupiah", u"Rp", 2)
CURRENCY["ILS"] = Currency("ILS", 376, "New Israeli Sheqel", u"₪", 2)
CURRENCY["INR"] = Currency("INR", 356, "Indian Rupee", u"₹", 2)
CURRENCY["IQD"] = Currency("IQD", 368, "Iraqi Dinar", u"", 3)
CURRENCY["IRR"] = Currency("IRR", 364, "Iranian Rial", u"﷼", 2)
CURRENCY["ISK"] = Currency("ISK", 352, "Iceland Krona", u"kr", 0)
CURRENCY["JMD"] = Currency("JMD", 388, "Jamaican Dollar", u"J$", 2)
CURRENCY["JOD"] = Currency("JOD", 400, "Jordanian Dinar", u"", 3)
CURRENCY["JPY"] = Currency("JPY", 392, "Yen", u"¥", 0)
CURRENCY["KES"] = Currency("KES", 404, "Kenyan Shilling", u"", 2)
CURRENCY["KGS"] = Currency("KGS", 417, "Som", u"лв", 2)
CURRENCY["KHR"] = Currency("KHR", 116, "Riel", u"៛", 2)
CURRENCY["KMF"] = Currency("KMF", 174, "Comoro Franc", u"", 0)
CURRENCY["KPW"] = Currency("KPW", 408, "North Korean Won", u"₩", 2)
CURRENCY["KRW"] = Currency("KRW", 410, "Won", u"₩", 0)
CURRENCY["KWD"] = Currency("KWD", 414, "Kuwaiti Dinar", u"", 3)
CURRENCY["KYD"] = Currency("KYD", 136, "Cayman Islands Dollar", u"$", 2)
CURRENCY["KZT"] = Currency("KZT", 398, "Tenge", u"лв", 2)
CURRENCY["LAK"] = Currency("LAK", 418, "Kip", u"₭", 2)
CURRENCY["LBP"] = Currency("LBP", 422, "Lebanese Pound", u"£", 2)
CURRENCY["LKR"] = Currency("LKR", 144, "Sri Lanka Rupee", u"₨", 2)
CURRENCY["LRD"] = Currency("LRD", 430, "Liberian Dollar", u"$", 2)
CURRENCY["LSL"] = Currency("LSL", 426, "Loti", u"", 2)
CURRENCY["LTL"] = Currency("LTL", 440, "Lithuanian Litas", u"Lt", 2)
CURRENCY["LYD"] = Currency("LYD", 434, "Libyan Dinar", u"", 3)
CURRENCY["MAD"] = Currency("MAD", 504, "Moroccan Dirham", u"", 2)
CURRENCY["MDL"] = Currency("MDL", 498, "Moldovan Leu", u"", 2)
CURRENCY["MGA"] = Currency("MGA", 969, "Malagasy Ariary", u"", 2)
CURRENCY["MKD"] = Currency("MKD", 807, "Denar", u"ден", 2)
CURRENCY["MMK"] = Currency("MMK", 104, "Kyat", u"", 2)
CURRENCY["MNT"] = Currency("MNT", 496, "Tugrik", u"₮", 2)
CURRENCY["MOP"] = Currency("MOP", 446, "Pataca", u"", 2)
CURRENCY["MRO"] = Currency("MRO", 478, "Ouguiya", u"", 2)
CURRENCY["MUR"] = Currency("MUR", 480, "Mauritius Rupee", u"₨", 2)
CURRENCY["MVR"] = Currency("MVR", 462, "Rufiyaa", u"", 2)
CURRENCY["MWK"] = Currency("MWK", 454, "Kwacha", u"", 2)
CURRENCY["MXN"] = Currency("MXN", 484, "Mexican Peso", u"$", 2)
CURRENCY["MXV"] = Currency("MXV", 979, "Mexican Unidad de Inversion (UDI)", u"", 2)
CURRENCY["MYR"] = Currency("MYR", 458, "Malaysian Ringgit", u"RM", 2)
CURRENCY["MZN"] = Currency("MZN", 943, "Mozambique Metical", u"MT", 2)
CURRENCY["NAD"] = Currency("NAD", 516, "Namibia Dollar", u"$", 2)
CURRENCY["NGN"] = Currency("NGN", 566, "Naira", u"₦", 2)
CURRENCY["NIO"] = Currency("NIO", 558, "Cordoba Oro", u"C$", 2)
CURRENCY["NOK"] = Currency("NOK", 578, "Norwegian Krone", u"kr", 2)
CURRENCY["NPR"] = Currency("NPR", 524, "Nepalese Rupee", u"₨", 2)
CURRENCY["NZD"] = Currency("NZD", 554, "New Zealand Dollar", u"$", 2)
CURRENCY["OMR"] = Currency("OMR", 512, "Rial Omani", u"﷼", 3)
CURRENCY["PAB"] = Currency("PAB", 590, "Balboa", u"B/.", 2)
CURRENCY["PEN"] = Currency("PEN", 604, "Nuevo Sol", u"S/.", 2)
CURRENCY["PGK"] = Currency("PGK", 598, "Kina", u"", 2)
CURRENCY["PHP"] = Currency("PHP", 608, "Philippine Peso", u"₱", 2)
CURRENCY["PKR"] = Currency("PKR", 586, "Pakistan Rupee", u"₨", 2)
CURRENCY["PLN"] = Currency("PLN", 985, "Zloty", u"zł", 2)
CURRENCY["PYG"] = Currency("PYG", 600, "Guarani", u"Gs", 0)
CURRENCY["QAR"] = Currency("QAR", 634, "Qatari Rial", u"﷼", 2)
CURRENCY["RON"] = Currency("RON", 946, "New Romanian Leu", u"lei", 2)
CURRENCY["RSD"] = Currency("RSD", 941, "Serbian Dinar", u"Дин.", 2)
CURRENCY["RUB"] = Currency("RUB", 643, "Russian Ruble", u"руб", 2)
CURRENCY["RWF"] = Currency("RWF", 646, "Rwanda Franc", u"", 0)
CURRENCY["SAR"] = Currency("SAR", 682, "Saudi Riyal", u"", 2)
CURRENCY["SBD"] = Currency("SBD", 90, "Solomon Islands Dollar", u"$", 2)
CURRENCY["SCR"] = Currency("SCR", 690, "Seychelles Rupee", u"₨", 2)
CURRENCY["SDG"] = Currency("SDG", 938, "Sudanese Pound", u"", 2)
CURRENCY["SEK"] = Currency("SEK", 752, "Swedish Krona", u"kr", 2)
CURRENCY["SGD"] = Currency("SGD", 702, "Singapore Dollar", u"$", 2)
CURRENCY["SHP"] = Currency("SHP", 654, "Saint Helena Pound", u"£", 2)
CURRENCY["SLL"] = Currency("SLL", 694, "Leone", u"", 2)
CURRENCY["SOS"] = Currency("SOS", 706, "Somali Shilling", u"S", 2)
CURRENCY["SRD"] = Currency("SRD", 968, "Surinam Dollar", u"$", 2)
CURRENCY["SSP"] = Currency("SSP", 728, "South Sudanese Pound", u"", 2)
CURRENCY["STD"] = Currency("STD", 678, "Dobra", u"", 2)
CURRENCY["SVC"] = Currency("SVC", 222, "El Salvador Colon", u"$", 2)
CURRENCY["SYP"] = Currency("SYP", 760, "Syrian Pound", u"£", 2)
CURRENCY["SZL"] = Currency("SZL", 748, "Lilangeni", u"", 2)
CURRENCY["THB"] = Currency("THB", 764, "Baht", u"฿", 2)
CURRENCY["TJS"] = Currency("TJS", 972, "Somoni", u"", 2)
CURRENCY["TMT"] = Currency("TMT", 934, "Turkmenistan New Manat", u"", 2)
CURRENCY["TND"] = Currency("TND", 788, "Tunisian Dinar", u"", 3)
CURRENCY["TOP"] = Currency("TOP", 776, "Pa’anga", u"", 2)
CURRENCY["TRY"] = Currency("TRY", 949, "Turkish Lira", u"₤", 2)
CURRENCY["TTD"] = Currency("TTD", 780, "Trinidad and Tobago Dollar", u"TT$", 2)
CURRENCY["TWD"] = Currency("TWD", 901, "New Taiwan Dollar", u"NT$", 2)
CURRENCY["TZS"] = Currency("TZS", 834, "Tanzanian Shilling", u"", 2)
CURRENCY["UAH"] = Currency("UAH", 980, "Hryvnia", u"₴", 2)
CURRENCY["UGX"] = Currency("UGX", 800, "Uganda Shilling", u"", 0)
CURRENCY["USD"] = Currency("USD", 840, "US Dollar", u"$", 2)
CURRENCY["USN"] = Currency("USN", 997, "US Dollar (Next day)", u"", 2)
CURRENCY["UYI"] = Currency(
    "UYI", 940, "Uruguay Peso en Unidades Indexadas (URUIURUI)", u"", 0
)
CURRENCY["UYU"] = Currency("UYU", 858, "Peso Uruguayo", u"$U", 2)
CURRENCY["UZS"] = Currency("UZS", 860, "Uzbekistan Sum", u"лв", 2)
CURRENCY["VEF"] = Currency("VEF", 937, "Bolivar", u"Bs", 2)
CURRENCY["VND"] = Currency("VND", 704, "Dong", u"₫", 0)
CURRENCY["VUV"] = Currency("VUV", 548, "Vatu", u"", 0)
CURRENCY["WST"] = Currency("WST", 882, "Tala", u"", 2)
CURRENCY["XAF"] = Currency("XAF", 950, "CFA Franc BEAC", u"", 0)
CURRENCY["XCD"] = Currency("XCD", 951, "East Caribbean Dollar", u"$", 2)
CURRENCY["XOF"] = Currency("XOF", 952, "CFA Franc BCEAO", u"", 0)
CURRENCY["XPF"] = Currency("XPF", 953, "CFP Franc", u"", 0)
CURRENCY["XTS"] = Currency("XTS", 963, "Test currency", u"", 0)
CURRENCY["YER"] = Currency("YER", 886, "Yemeni Rial", u"﷼", 2)
CURRENCY["ZAR"] = Currency("ZAR", 710, "Rand", u"S", 2)
CURRENCY["ZMW"] = Currency("ZMW", 967, "Zambian Kwacha", u"", 2)
CURRENCY["ZWL"] = Currency("ZWL", 932, "Zimbabwe Dollar", u"Z$", 2)

# Additional special case as BTC is not a "real" currency.
CURRENCY["BTC"] = Currency("BTC", 0, "Bitcoin", u"฿", 4)



try:
    from pint import UnitRegistry
except ImportError:
    raise ImportError(
        "The pint package is not installed, please install this library to use quantity fields."
    )

# Common unit registry
registry = UnitRegistry()



import arrow
import datetime
import pytest

from odin.exceptions import ValidationError

try:
    from odin.contrib.arrow.fields import ArrowField
except ImportError:
    pass  # Arrow is not available
else:

    class TestArrowField:
        @pytest.mark.parametrize(
            "kwargs, value, expected",
            (
                ({"null": True}, None, None),
                ({}, "2013-11-24", arrow.Arrow(2013, 11, 24)),
                ({}, "2013-11-24T18:43:00.000", arrow.Arrow(2013, 11, 24, 18, 43)),
                ({}, "2013-11-24T18:43:00.000Z", arrow.Arrow(2013, 11, 24, 18, 43)),
                (
                    {},
                    arrow.Arrow(2013, 11, 24, 18, 43),
                    arrow.Arrow(2013, 11, 24, 18, 43),
                ),
                (
                    {},
                    datetime.datetime(2013, 11, 24, 18, 43),
                    arrow.Arrow(2013, 11, 24, 18, 43),
                ),
                (
                    {},
                    datetime.datetime(
                        2013, 11, 24, 18, 43, 0, 0, datetime.timezone.utc
                    ),
                    arrow.Arrow(2013, 11, 24, 18, 43),
                ),
            ),
        )
        def test_arrowfield__clean__valid(self, kwargs, value, expected):
            target = ArrowField(**kwargs)
            actual = target.clean(value)

            assert actual == expected

        @pytest.mark.parametrize(
            "kwargs, value",
            (
                ({}, None),
                ({}, 123),
                ({}, "2013-11-"),
            ),
        )
        def test_arrowfield__clean__invalid(self, kwargs, value):
            with pytest.raises(ValidationError):
                ArrowField(**kwargs).clean(value)



"""
Resource Proxy
~~~~~~~~~~~~~~

The resource proxy is an object that provides an alternate interface to a shadowed Resource.

This could be as a means of providing a summary object, or for adding additional properties.

"""
from typing import List, Union, Optional

from odin import registration
from odin.bases import TypedResourceIterable
from odin.resources import ResourceOptions, ResourceBase
from odin.utils import getmeta, filter_fields, lazy_property

EMPTY = tuple()


class FieldProxyDescriptor:
    """
    Descriptor to proxy field to underlying resource.
    """

    __slots__ = ("attname", "readonly")

    def __init__(self, readonly=False):
        self.attname: Optional[str] = None
        self.readonly = readonly

    def __get__(self, instance, owner):
        shadow = getattr(instance, "_shadow")
        return getattr(shadow, self.attname)

    def __set__(self, instance, value):
        if self.readonly:
            raise AttributeError("can't set attribute")
        shadow = getattr(instance, "_shadow")
        return setattr(shadow, self.attname, value)

    def contribute_to_class(self, cls, name):
        self.attname = name
        setattr(cls, name, self)


class ResourceProxyOptions(ResourceOptions):
    META_OPTION_NAMES = (
        "include",
        "exclude",
        "readonly",
        # Overrides
        "name",
        "name_space",
        "namespace",
        "verbose_name",
        "verbose_name_plural",
        "doc_group",
        "type_field",
        "key_field_name",
        "key_field_names",
        "field_sorting",
    )

    def __init__(self, meta):
        super(ResourceProxyOptions, self).__init__(meta)
        self.resource = None
        self.shadow = None
        self.include = []
        self.exclude = []
        self.readonly = []

    def __repr__(self):
        return "<Proxy of {!r}>".format(getmeta(self.resource))

    def contribute_to_class(self, cls, _):
        cls._meta = self
        cls_name = cls.__name__
        self.name = cls_name
        self.class_name = f"{cls.__module__}.{cls_name}"

        # Get and filter meta attributes
        meta_attrs = self.meta.__dict__.copy()
        for name in self.meta.__dict__:
            if name.startswith("_"):
                del meta_attrs[name]

        # Get the required resource object
        self.resource = meta_attrs.pop("resource", None)
        if not self.resource:
            raise AttributeError("`resource` has not been defined.")
        self.shadow = shadow = getmeta(self.resource)

        # Extract all meta options and fetch from shadow if not defined
        proxy_attrs = {
            "name": cls_name,
            "verbose_name": cls_name.replace("_", " ").strip("_ "),
        }
        for attr_name in self.META_OPTION_NAMES:
            if attr_name in meta_attrs:
                value = meta_attrs.pop(attr_name)

                if attr_name == "verbose_name":
                    # If defined generate pluralised form base on this name.
                    if "verbose_name_plural" not in proxy_attrs:
                        proxy_attrs["verbose_name_plural"] = value + "s"

                elif attr_name == "namespace":
                    # Allow meta to be defined as namespace
                    attr_name = "name_space"

                elif attr_name == "key_field_name":
                    # Remap to key_field names
                    attr_name = "key_field_names"
                    value = [value]

                proxy_attrs[attr_name] = value

            elif hasattr(shadow, attr_name):
                proxy_attrs.setdefault(attr_name, getattr(shadow, attr_name))

        # Any leftover attributes must be invalid.
        if meta_attrs != {}:
            raise TypeError(
                f"'class Meta' got invalid attribute(s): {','.join(meta_attrs.keys())}"
            )
        del self.meta

        # Apply to self
        for attr_name, value in proxy_attrs.items():
            setattr(self, attr_name, value)

    @lazy_property
    def readonly_fields(self):
        """
        Fields that can only be read from.
        """
        return tuple(f for f in self.fields if f.attname in self.readonly)

    @lazy_property
    def init_fields(self):
        """
        Fields used in the resource init
        """
        return tuple(f for f in self.fields if f.attname not in self.readonly)


class ResourceProxyType(type):
    meta_options = ResourceProxyOptions

    def __new__(mcs, name, bases, attrs):
        super_new = super(ResourceProxyType, mcs).__new__

        # attrs will never be empty for classes declared in the standard way
        # (ie. with the `class` keyword). This is quite robust.
        if name == "NewBase" and attrs == {}:
            return super_new(mcs, name, bases, attrs)

        parents = [
            b
            for b in bases
            if isinstance(b, ResourceProxyType)
            and not (b.__name__ == "NewBase" and b.__mro__ == (b, object))
        ]
        if not parents:
            # If this isn't a subclass of Resource, don't do anything special.
            return super_new(mcs, name, bases, attrs)

        # Create the class.
        module = attrs.pop("__module__")
        new_class = super_new(mcs, name, bases, {"__module__": module})
        attr_meta = attrs.pop("Meta", None)
        if not attr_meta:
            meta = getattr(new_class, "Meta", None)
        else:
            meta = attr_meta

        new_meta = mcs.meta_options(meta)
        new_class.add_to_class("_meta", new_meta)

        # Add all attributes to the class.
        for obj_name, obj in attrs.items():
            new_class.add_to_class(obj_name, obj)

        # Determine which fields will be shadowed.
        field_names, readonly = filter_fields(
            new_meta.shadow.field_map,
            new_meta.include,
            new_meta.exclude,
            new_meta.readonly,
        )

        # Map field names
        new_meta.fields = [
            f for f in new_meta.shadow.fields if f.attname in field_names
        ]

        # Sort the fields
        if new_meta.field_sorting:
            if callable(new_meta.field_sorting):
                new_meta.fields = new_meta.field_sorting(new_meta.fields)
            else:
                new_meta.fields = sorted(new_meta.fields, key=hash)

        # Generate field descriptors
        for field in new_meta.fields:
            new_class.add_to_class(
                field.attname, FieldProxyDescriptor(field in new_meta.readonly_fields)
            )

        # If a key_field is defined ensure it exists
        if new_meta.key_field_names:
            for field_name in new_meta.key_field_names:
                if field_name not in new_meta.field_map:
                    raise AttributeError(
                        "Key field {!r} is not exist on this resource.".format(
                            field_name
                        )
                    )

        # Register resource
        registration.register_resources(new_class)

        # Because of the way imports happen (recursively), we may or may not be
        # the first time this model tries to register with the framework. There
        # should only be one class for each model, so we always return the
        # registered version.
        return registration.get_resource(new_meta.resource_name)

    def add_to_class(cls, name, value):
        if hasattr(value, "contribute_to_class"):
            value.contribute_to_class(cls, name)
        else:
            setattr(cls, name, value)


class ResourceProxyIter(TypedResourceIterable):
    """
    Returned by `ResourceProxy.proxy` if an iterable is supplied.
    """

    def __init__(self, objects, resource_type):
        # type: (Union[List[ResourceBase]], ResourceProxyBase) -> None
        super(ResourceProxyIter, self).__init__(resource_type)
        self.objects = objects

    def __iter__(self):
        resource_type = self.resource_type
        for obj in self.objects:
            yield resource_type.proxy(obj)


class ResourceProxyBase(ResourceBase):
    """
    Base proxy class
    """

    @classmethod
    def proxy(cls, resource):
        if hasattr(resource, "__iter__"):
            return ResourceProxyIter(resource, cls)
        elif isinstance(resource, ResourceBase):
            return cls(__resource=resource)
        raise TypeError("Only resource types can be proxied.")

    def __init__(self, *args, **kwargs):
        meta = getmeta(self)

        # Get shadowed resource if supplied
        shadow = kwargs.pop("__resource", None)
        if shadow is None:
            # Create a new instance
            self._shadow = meta.resource()
            super(ResourceProxyBase, self).__init__(*args, **kwargs)
        else:
            self._shadow = shadow

    def get_shadow(self):
        return self._shadow


class ResourceProxy(ResourceProxyBase, metaclass=ResourceProxyType):
    """
    Proxy for a Resources that allow a filtered set of fields to be made
    available and updated.
    """



import pytest

from odin import validators
from odin.exceptions import ValidationError


class TestValidator:
    def test_regex_validator(self):
        target = validators.RegexValidator(
            r"^[a-z]{3}$", "Please enter 3 alpha characters.", "match_chars"
        )

        assert "Please enter 3 alpha characters." == target.message
        assert "match_chars" == target.code

        target("abc")
        target("cba")
        pytest.raises(ValidationError, target, "a")
        pytest.raises(ValidationError, target, "abcd")
        pytest.raises(ValidationError, target, "123")

    def test_url(self):
        validators.validate_url("http://www.djangoproject.com/")
        validators.validate_url("http://localhost/")
        validators.validate_url("http://example.com/")
        validators.validate_url("http://www.example.com/")
        validators.validate_url("http://www.example.com:8000/test")
        validators.validate_url("http://valid-with-hyphens.com/")
        validators.validate_url("http://subdomain.example.com/")
        validators.validate_url("http://200.8.9.10/")
        validators.validate_url("http://200.8.9.10:8000/test")
        validators.validate_url("http://valid-----hyphens.com/")
        validators.validate_url("http://example.com?something=value")
        validators.validate_url(
            "http://example.com/index.php?something=value&another=value2"
        )
        validators.validate_url("https://example.com/")
        validators.validate_url("ftp://example.com/")
        validators.validate_url("ftps://example.com/")
        validators.validate_url("http://savage.company/")

        pytest.raises(ValidationError, validators.validate_url, "foo")
        pytest.raises(ValidationError, validators.validate_url, "http://")
        pytest.raises(ValidationError, validators.validate_url, "http://example")
        pytest.raises(ValidationError, validators.validate_url, "http://example.")
        pytest.raises(ValidationError, validators.validate_url, "http://.com")
        pytest.raises(ValidationError, validators.validate_url, "http://invalid-.com")
        pytest.raises(ValidationError, validators.validate_url, "http://-invalid.com")
        pytest.raises(ValidationError, validators.validate_url, "http://inv-.alid-.com")
        pytest.raises(ValidationError, validators.validate_url, "http://inv-.-alid.com")

    def test_max_value_validator(self):
        target = validators.MaxValueValidator(10)

        target(10)
        target(1)
        target(-10)
        pytest.raises(ValidationError, target, 11)

    def test_min_value_validator(self):
        target = validators.MinValueValidator(10)

        target(10)
        pytest.raises(ValidationError, target, 1)
        pytest.raises(ValidationError, target, -10)
        target(11)

    def test_length_validator(self):
        target = validators.LengthValidator(10)
        target("1234567890")
        pytest.raises(ValidationError, target, "123456789")
        pytest.raises(ValidationError, target, "12345678901")

    def test_max_length_validator(self):
        target = validators.MaxLengthValidator(10)

        target("123457890")
        target("12345")
        target("")
        pytest.raises(ValidationError, target, "12345678901")

    def test_min_length_validator(self):
        target = validators.MinLengthValidator(10)

        pytest.raises(ValidationError, target, "123457890")
        pytest.raises(ValidationError, target, "12345")
        pytest.raises(ValidationError, target, "")
        target("12345678901")


class TestSimpleValidator:
    def test_method(self):
        def reflect(v):
            return v

        validator = validators.simple_validator(reflect)

        pytest.raises(ValidationError, validator, False)
        validator(True)

    def test_decorator(self):
        @validators.simple_validator()
        def reflect_validator(v):
            return v

        pytest.raises(ValidationError, reflect_validator, False)
        reflect_validator(True)

    def test_str__no_doc_text(self):
        @validators.simple_validator()
        def reflect_validator(v):
            return v

        assert str(reflect_validator) == "reflect_validator"

    def test_str__doc_text(self):
        @validators.simple_validator()
        def reflect_validator(v):
            """
            Reflect Validator
            """
            return v

        assert str(reflect_validator) == "Reflect Validator"


@pytest.mark.parametrize(
    "value",
    (
        "192.168.0.1",
        "1.1.1.1",
        "255.255.255.255",
    ),
)
def test_ipv4_address_valid(value):
    validators.validate_ipv4_address(value)


@pytest.mark.parametrize(
    "value",
    (
        ".2.3.4",
        "1.2.3",
        "1.2.3.4.5",
        "300.0.0.1",
        "ff.ff.ff.ff",
        "abcd",
        "",
        "::ff:ff:ff:ff",
    ),
)
def test_ipv4_address_invalid(value):
    with pytest.raises(ValidationError):
        validators.validate_ipv4_address(value)


@pytest.mark.parametrize(
    "value",
    (
        "::",  # Localhost
        "::1",  # Default unicast
        "::ffff:0:0",  # IPv4 Mapped address
        "2002::",  # 6to4
        "2001::",  # Teredo tunneling
        "fc00::",  # Unique local address
        "ffff::192.168.0.1",
        "1:2:3:4:5:6:7:8",
        "dead:beef:dead:beef:dead:beef:dead:beef",
    ),
)
def test_ipv6_address_valid(value):
    validators.validate_ipv6_address(value)


@pytest.mark.parametrize(
    "value",
    (
        "192.168.0.1",
        "1:2:3:4:5:6:7:8:9",
        "1234:5678:9abc:defg::",
        "2001::12345",
        "::ffff::",
        "ff:::0",
        ":2:3:4:5:6:7:8",
        "1:2:3:4:5:6:7:",
        ":2:3:4:5:6:7:",
        "1:2:3:4:5:192.168.0",
        "1:2:3:4:256.256.256.256",
        "192.168.0.1::ff",
        "zzzz::0",
    ),
)
def test_ipv6_address_invalid(value):
    with pytest.raises(ValidationError):
        validators.validate_ipv6_address(value)


@pytest.mark.parametrize(
    "value",
    (
        "192.168.0.1",
        "1:2:3:4:5:6:7:8",
    ),
)
def test_validate_ipv46_address_valid(value):
    validators.validate_ipv46_address(value)


@pytest.mark.parametrize(
    "value",
    (
        "192.168.0.",
        "1:2:3:4:5:6:7:8:9",
    ),
)
def test_validate_ipv46_address_invalid(value):
    with pytest.raises(ValidationError):
        validators.validate_ipv46_address(value)


@pytest.mark.parametrize(
    "value kwargs".split(),
    (
        ("foo@bar.com", None),
        ("FOO@BAR.COM", None),
        ("foo@localhost", None),
        # New release domain names
        ("foo@example.company", None),
        # IPv4 based email
        ("foo@[127.0.0.1]", None),
        # IPv6 based email
        ("foo@[::1]", None),
    ),
)
def test_validate_email_valid(value, kwargs):
    kwargs = kwargs or {}
    validators.EmailValidator(**kwargs)(value)


@pytest.mark.parametrize(
    "value kwargs".split(),
    (
        ("foo", None),
        (".foo@bar.com", None),
        ("foo@corpau", None),
        ("foo@localhost", {"whitelist": ""}),
        ("foo@{0}.com".format("a" * 64), None),
        ("foo@[::ff::0]", None),
    ),
)
def test_validate_email_invalid(value, kwargs):
    kwargs = kwargs or {}
    with pytest.raises(ValidationError):
        validators.EmailValidator(**kwargs)(value)



import pytest
import os

from odin import filtering
from odin.traversal import TraversalPath
from odin.codecs import json_codec


# Open library fixture
with open(os.path.join(os.path.dirname(__file__), "fixtures", "library.json")) as f:
    library = json_codec.load(f)

first_book = TraversalPath.parse("books[0]").get_value(library)


class TestFilter:
    def test_matches_top_level_resource(self):
        flt = filtering.Equal("title", "Consider Phlebas")
        assert flt(first_book)

    def test_matches_multiple_matches_resource(self):
        # Less than fiction $20 with a single author and not fantasy
        flt = filtering.And(
            filtering.Equal("fiction", True),
            filtering.LessThan("rrp", 20),
            filtering.Equal("authors", 1, len),
            filtering.NotEqual("genre", "fantasy"),
        )
        assert flt(first_book)

    def test_matches_sub_resource(self):
        flt = filtering.Equal("publisher.name", "Macmillan")
        assert flt(first_book)

    def test_find_sub_resources(self):
        flt = filtering.Equal("fiction", True)
        # Some fiction?
        assert flt.any(library.books)
        # All fiction
        assert not flt.all(library.books)

    def test_resource_field_does_not_exist(self):
        flt = filtering.Equal("title", "Public Library")
        assert not flt(library)


class TestFilterChain:
    def test_empty_description(self):
        flt = filtering.And()
        assert "" == str(flt)

    def test_description(self):
        flt = filtering.And(
            filtering.Equal("fiction", True),
            filtering.LessThan("rrp", 20),
        )
        assert "(fiction == True AND rrp < 20)" == str(flt)

    def test_combining_same_type(self):
        flt1 = filtering.And(
            filtering.Equal("fiction", True),
            filtering.LessThan("rrp", 20),
        )
        flt2 = filtering.And(
            filtering.Equal("authors", 1, len),
            filtering.NotEqual("genre", "fantasy"),
        )
        flt = flt1 + flt2

        assert (
            "(fiction == True AND rrp < 20 AND len(authors) == 1 AND genre != 'fantasy')"
            == str(flt)
        )
        assert len(flt) == 4

    def test_combine_with_comparison(self):
        flt1 = filtering.And(
            filtering.Equal("fiction", True),
            filtering.LessThan("rrp", 20),
        )
        flt = flt1 + filtering.Equal("authors", 1, len)

        assert "(fiction == True AND rrp < 20 AND len(authors) == 1)" == str(flt)

    def test_combine_with_other_type(self):
        flt1 = filtering.And(
            filtering.Equal("fiction", True),
            filtering.LessThan("rrp", 20),
        )
        pytest.raises(TypeError, lambda: flt1 + "abc")


class FilterComparisonTest(filtering.FilterComparison):
    operator_symbols = ["%"]


class TestFilterComparison:
    def test_description(self):
        comp = FilterComparisonTest("foo", "bar")
        assert "foo % 'bar'" == str(comp)

        comp = FilterComparisonTest("foo", 42)
        assert "foo % 42" == str(comp)

        comp = FilterComparisonTest("foo", "bar", any)
        assert "any(foo) % 'bar'" == str(comp)

    def test_equal(self):
        comp = filtering.Equal("foo", "bar")
        assert "foo == 'bar'" == str(comp)
        assert comp.compare_operator("bar", "bar")

    def test_not_equal(self):
        comp = filtering.NotEqual("foo", "bar")
        assert "foo != 'bar'" == str(comp)
        assert comp.compare_operator("bar", "eek")

    def test_less_than(self):
        comp = filtering.LessThan("foo", 5)
        assert "foo < 5" == str(comp)
        assert comp.compare_operator(4, 5)
        assert not comp.compare_operator(5, 5)
        assert not comp.compare_operator(6, 5)

    def test_less_than_or_equal(self):
        comp = filtering.LessThanOrEqual("foo", 5)
        assert "foo <= 5" == str(comp)
        assert comp.compare_operator(4, 5)
        assert comp.compare_operator(5, 5)
        assert not comp.compare_operator(6, 5)

    def test_greater_than(self):
        comp = filtering.GreaterThan("foo", 5)
        assert "foo > 5" == str(comp)
        assert not comp.compare_operator(4, 5)
        assert not comp.compare_operator(5, 5)
        assert comp.compare_operator(6, 5)

    def test_greater_than_or_equal(self):
        comp = filtering.GreaterThanOrEqual("foo", 5)
        assert "foo >= 5" == str(comp)
        assert not comp.compare_operator(4, 5)
        assert comp.compare_operator(5, 5)
        assert comp.compare_operator(6, 5)



import pytest
import datetime

from odin.exceptions import MappingSetupError, MappingExecutionError
from odin.fields import NotProvided
from odin.mapping import MappingResult, FieldMapping
from odin.mapping.helpers import MapDictAs, MapListOf, NoOpMapper
from .resources import *


class SimpleFromResource(odin.Resource):
    title = odin.StringField()


class SimpleToResource(odin.Resource):
    title = odin.StringField()
    title_count = odin.StringField()


class FakeToResource(odin.Resource):
    title = odin.StringField()
    name = odin.StringField()


class SimpleFromTo(odin.Mapping):
    from_obj = SimpleFromResource
    to_obj = SimpleToResource

    @odin.map_field(from_field="title")
    def title_count(self, value):
        if self.in_loop:
            return "%s: %s" % (self.loop_idx, value)
        else:
            return value


def assertMappingEquivalent(a, b):
    flat_a = sorted(str(i) for i in a)
    flat_b = sorted(str(i) for i in b)
    assert flat_a == flat_b


class TestMappingBase:
    maxDiff = None

    def test_full_mapping(self):
        assertMappingEquivalent(
            [
                FieldMapping(
                    ("from_field1",), None, ("to_field1",), False, False, False
                ),
                FieldMapping(
                    ("from_field2",), int, ("to_field2",), False, False, False
                ),
                FieldMapping(
                    ("from_field3", "from_field4"),
                    sum_fields,
                    ("to_field3",),
                    False,
                    False,
                    False,
                ),
                FieldMapping(
                    ("from_field1",), None, ("same_but_different",), False, False, False
                ),
                FieldMapping(
                    ("from_field_c1", "from_field_c2", "from_field_c3"),
                    "multi_to_one",
                    ("to_field_c1",),
                    False,
                    False,
                    False,
                ),
                FieldMapping(
                    ("from_field_c4",),
                    "one_to_multi",
                    ("to_field_c2", "to_field_c3"),
                    False,
                    False,
                    False,
                ),
                FieldMapping(
                    ("not_auto_c5",),
                    "not_auto_c5",
                    ("not_auto_c5",),
                    False,
                    False,
                    False,
                ),
                FieldMapping(
                    ("comma_separated_string",),
                    "comma_separated_string",
                    ("array_string",),
                    True,
                    False,
                    False,
                ),
                FieldMapping(
                    None, "assigned_field", ("assigned_field",), False, False, False
                ),
                FieldMapping(
                    ("count",), "default_action", ("count",), False, False, False
                ),
                FieldMapping(
                    ("child",), MapDictAs(NoOpMapper), ("child",), False, True, False
                ),
                FieldMapping(
                    ("children",),
                    MapListOf(NoOpMapper),
                    ("children",),
                    False,
                    True,
                    False,
                ),
                FieldMapping(
                    ("title",), "default_action", ("title",), False, False, False
                ),
            ],
            FromToMapping._mapping_rules,
        )

    def test_map(self):
        from_resource = FromResource(
            # Auto matched
            title="Foo",
            count="42",
            # Excluded
            excluded1=123.4,
            # Mappings
            from_field1="abc",
            from_field2="62",
            from_field3=44,
            from_field4=25,
            same_but_different="def",
            # Custom mappings
            from_field_c1="foo",
            from_field_c2="bar",
            from_field_c3="eek",
            from_field_c4="first-second-third",
            not_auto_c5="do something",
            comma_separated_string="foo,bar,eek",
            child=ChildResource(name="foo"),
            children=[ChildResource(name="foo"), ChildResource(name="bar")],
        )

        to_resource = from_resource.convert_to(ToResource)

        assert "Foo" == to_resource.title
        assert "42" == to_resource.count
        assert None is to_resource.excluded1
        assert "abc" == to_resource.to_field1
        assert 62 == to_resource.to_field2
        assert 69 == to_resource.to_field3
        assert "abc" == to_resource.same_but_different
        assert "foo-bar-eek" == to_resource.to_field_c1
        assert "first" == to_resource.to_field_c2
        assert "second-third" == to_resource.to_field_c3
        assert "DO SOMETHING" == to_resource.not_auto_c5
        assert ["foo", "bar", "eek"] == to_resource.array_string

    def test_missing_from_resource(self):
        with pytest.raises(MappingSetupError):

            class _(odin.Mapping):
                to_obj = FakeToResource

    def test_missing_to_resource(self):
        with pytest.raises(MappingSetupError):

            class _(odin.Mapping):
                from_obj = SimpleFromResource

    def test_unknown_from_field_mappings(self):
        with pytest.raises(MappingSetupError):

            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = (("unknown_field", None, "title"),)

    def test_unknown_from_field_mappings_multiple(self):
        with pytest.raises(MappingSetupError):

            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = ((("title", "unknown_field"), None, "title"),)

    def test_unknown_from_field_custom(self):
        with pytest.raises(MappingSetupError):

            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                @odin.map_field(from_field="unknown_field", to_field="title")
                def multi_to_one(self, *fields):
                    pass

    def test_bad_action_not_callable(self):
        with pytest.raises(MappingSetupError):

            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = (("title", 123, "title"),)

    def test_bad_action_not_defined(self):
        with pytest.raises(MappingSetupError):

            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = (("title", "do_transform", "title"),)

    def test_bad_action_defined_not_callable(self):
        with pytest.raises(MappingSetupError):

            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                do_transform = 123

                mappings = (("title", "do_transform", "title"),)

    def test_unknown_to_field_mappings(self):
        with pytest.raises(MappingSetupError):

            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = (("title", None, "unknown_field"),)

    def test_unknown_to_field_mappings_multiple(self):
        with pytest.raises(MappingSetupError):

            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = (("title", None, ("title", "unknown_field")),)

    def test_unknown_to_field_custom(self):
        with pytest.raises(MappingSetupError):

            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                @odin.map_field(from_field="title", to_field="unknown_field")
                def multi_to_one(self, *fields):
                    pass

    def test_bad_mapping(self):
        with pytest.raises(MappingSetupError):

            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = ("i_forgot", None, "tuples")

    def test_invalid_list_to_multiple_mapping(self):
        with pytest.raises(MappingSetupError) as cm:

            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = (("title", None, ("title", "name"), True, False, False),)

        assert "specifies a to_list mapping, these can only be applied to a" in str(
            cm.value
        )

    def test_assignment_with_no_action(self):
        with pytest.raises(MappingSetupError) as cm:

            class _(odin.Mapping):
                from_obj = SimpleFromResource
                to_obj = FakeToResource

                mappings = ((None, None, ("title", "name"), False, False, False),)

        assert "No action supplied for " in str(cm.value)

    def test_from_field_no_field_resolver(self):
        with pytest.raises(MappingSetupError) as cm:

            class _(odin.Mapping):
                from_obj = object
                to_obj = FakeToResource

        assert r"`from_obj`" in str(cm.value)
        assert r"does not have an attribute resolver defined." in str(cm.value)

    def test_to_field_no_field_resolver(self):
        with pytest.raises(MappingSetupError) as cm:

            class _(odin.Mapping):
                from_obj = FakeToResource
                to_obj = object

        assert r"`to_obj`" in str(cm.value)
        assert r"does not have an attribute resolver defined." in str(cm.value)


class TestExecuteMapping:
    def test_not_valid_from_resource(self):
        pytest.raises(TypeError, FromToMapping, SimpleFromResource())

    def test_invalid_from_value_count(self):
        with pytest.raises(MappingExecutionError):
            target = FromToMapping(FromResource())
            target._apply_rule(
                (
                    ("from_field_c1", "from_field_c2"),
                    "one_to_multi",
                    ("title"),
                    False,
                    False,
                    False,
                )
            )

    def test_invalid_to_value_count(self):
        with pytest.raises(MappingExecutionError):
            target = FromToMapping(FromResource(title="Test"))
            target._apply_rule(
                (
                    ("title",),
                    "multi_to_one",
                    ("from_field_c1", "from_field_c2"),
                    False,
                    False,
                    False,
                )
            )

    def test_apply_single_resource(self):
        f = SimpleFromResource(title="ABC")
        t = SimpleFromTo.apply(f)

        assert isinstance(t, SimpleToResource)
        assert "ABC" == t.title

    def test_apply_single_resource_with_context(self):
        f = SimpleFromResource(title="ABC")
        t = SimpleFromTo.apply(f)

        assert isinstance(t, SimpleToResource)
        assert "ABC" == t.title_count

    def test_apply_multiple_resources(self):
        from_resources = [
            SimpleFromResource(title="Foo"),
            SimpleFromResource(title="Bar"),
            SimpleFromResource(title="Eek"),
        ]

        to_resource_iter = SimpleFromTo.apply(from_resources)
        assert isinstance(to_resource_iter, MappingResult)
        assert ["Foo", "Bar", "Eek"] == [t.title for t in to_resource_iter]

    def test_apply_multiple_resources_with_context(self):
        from_resources = [
            SimpleFromResource(title="Foo"),
            SimpleFromResource(title="Bar"),
            SimpleFromResource(title="Eek"),
        ]

        to_resource_iter = SimpleFromTo.apply(from_resources)
        assert isinstance(to_resource_iter, MappingResult)
        assert ["0: Foo", "1: Bar", "2: Eek"] == [
            t.title_count for t in to_resource_iter
        ]


class TestResourceMappingHelpers:
    def test_update_existing(self):
        resource = OldBook(
            name="Bar",  # odin.StringField()
            num_pages=42,  # odin.IntegerField()
            price=12.50,  # odin.FloatField()
            genre="fantasy",  # odin.StringField()
            published=datetime.datetime(2015, 2, 26, 22, 7),  # odin.DateTimeField()
            author=Author(name="123"),  # odin.ObjectAs()
            publisher=Publisher(name="Eek"),  # odin.ObjectAs()
        )
        existing_resource = Book(
            title="Foo",  # odin.StringField()
            num_pages=12,  # odin.IntegerField()
            rrp=12.50,  # odin.FloatField()
            fiction=False,  # odin.BooleanField()
            genre="others",  # odin.StringField()
            published=[datetime.datetime.now()],  # odin.TypedArrayField()
        )

        resource.update_existing(existing_resource)

        assert "Bar" == existing_resource.title
        assert 42 == existing_resource.num_pages
        assert 12.50 == existing_resource.rrp
        assert not existing_resource.fiction
        assert "fantasy" == existing_resource.genre

    def test_update_filtered_existing(self):
        resource = OldBook(
            name="Bar",  # odin.StringField()
            num_pages=42,  # odin.IntegerField()
            price=12.50,  # odin.FloatField()
            genre="fantasy",  # odin.StringField()
            published=datetime.datetime(2015, 2, 26, 22, 7),  # odin.DateTimeField()
            author=Author(name="123"),  # odin.ObjectAs()
            publisher=Publisher(name="Eek"),  # odin.ObjectAs()
        )
        existing_resource = Book(
            title="Foo",  # odin.StringField()
            num_pages=12,  # odin.IntegerField()
            rrp=12.50,  # odin.FloatField()
            fiction=False,  # odin.BooleanField()
            genre="others",  # odin.StringField()
            published=[datetime.datetime.now()],  # odin.TypedArrayField()
        )

        resource.update_existing(existing_resource, ignore_fields=["num_pages"])

        assert "Bar" == existing_resource.title
        assert 12 == existing_resource.num_pages
        assert 12.50 == existing_resource.rrp
        assert not existing_resource.fiction
        assert "fantasy" == existing_resource.genre

    def test_ignore_not_provided(self):
        resource = OldBook(
            name="Bar",  # odin.StringField()
            num_pages=NotProvided,  # odin.IntegerField()
            price=12.50,  # odin.FloatField()
            genre=NotProvided,  # odin.StringField()
            published=datetime.datetime(2015, 2, 26, 22, 7),  # odin.DateTimeField()
            author=Author(name="123"),  # odin.ObjectAs()
            publisher=Publisher(name="Eek"),  # odin.ObjectAs()
        )
        existing_resource = Book(
            title="Foo",  # odin.StringField()
            num_pages=12,  # odin.IntegerField()
            rrp=12.50,  # odin.FloatField()
            fiction=False,  # odin.BooleanField()
            genre="others",  # odin.StringField()
            published=[datetime.datetime.now()],  # odin.TypedArrayField()
        )

        resource.update_existing(existing_resource, ignore_not_provided=True)

        assert "Bar" == existing_resource.title
        assert 12 == existing_resource.num_pages
        assert 12.50 == existing_resource.rrp
        assert not existing_resource.fiction
        assert "others" == existing_resource.genre


class ResourceA(odin.Resource):
    class Meta:
        abstract = True

    foo = odin.StringField()


class ResourceB(ResourceA):
    bar = odin.StringField()


class ResourceC(ResourceA):
    eek = odin.StringField()


class ResourceX(odin.Resource):
    class Meta:
        abstract = True

    foo = odin.StringField()


class ResourceY(ResourceX):
    bar = odin.StringField()


class ResourceZ(ResourceX):
    alt = odin.StringField()


class ResourceAToResourceX(odin.Mapping):
    from_obj = ResourceA
    to_obj = ResourceX

    @odin.map_field
    def foo(self, value):
        return "foo: %s" % value


class ResourceBToResourceY(ResourceAToResourceX):
    from_obj = ResourceB
    to_obj = ResourceY


class ResourceCToResourceZ(ResourceAToResourceX):
    from_obj = ResourceC
    to_obj = ResourceZ

    @odin.map_field(from_field="eek")
    def alt(self, value):
        return value


class TestSubClassMapping:
    """
    Test the concept of a sub class mapping ie

                Resource A          -->         Mapping A-X         -->           Resource X
                    |                                |                                |
               _____|_____                      _____|_____                      _____|_____
              /           \                    /           \                    /           \
             /            \                   /            \                   /            \
        Resource B    Resource C         Mapping B-Y   Mapping C-Z        Resource Y    Resource Z

    Define a mapping that can handle a list of *Resource A* and *Resource B* objects being mapped by an abstract mapping
    *Mapping A-X* to the corresponding *Resource Y* and *Resource Z* objects.

    """

    def test_abstract_resource_definitions(self):
        assert ["bar", "foo"] == [f.name for f in ResourceB._meta.fields]
        assert ["eek", "foo"] == [f.name for f in ResourceC._meta.fields]
        assert ["bar", "foo"] == [f.name for f in ResourceY._meta.fields]
        assert ["alt", "foo"] == [f.name for f in ResourceZ._meta.fields]

    def test_abstract_mapping_definitions(self):
        assertMappingEquivalent(
            [
                FieldMapping(("foo",), "foo", ("foo",), False, False, False),
                FieldMapping(("bar",), "default_action", ("bar",), False, False, False),
            ],
            ResourceBToResourceY._mapping_rules,
        )

        assertMappingEquivalent(
            [
                FieldMapping(("eek",), "alt", ("alt",), False, False, False),
                FieldMapping(("foo",), "foo", ("foo",), False, False, False),
            ],
            ResourceCToResourceZ._mapping_rules,
        )

    def test_abstract_mapping(self):
        source = [
            ResourceB(foo="1", bar="2"),
            ResourceC(foo="3", eek="4"),
            ResourceB(foo="5", bar="6"),
        ]
        result = ResourceAToResourceX.apply(source)

        assert isinstance(result[0], ResourceY)
        assert isinstance(result[1], ResourceZ)
        assert isinstance(result[2], ResourceY)

    def test_mapping_to_list(self):
        source = [
            ResourceB(foo="1", bar="2"),
            ResourceC(foo="3", eek="4"),
            ResourceB(foo="5", bar="6"),
        ]
        result = list(ResourceAToResourceX.apply(source))

        assert isinstance(result[0], ResourceY)
        assert isinstance(result[1], ResourceZ)
        assert isinstance(result[2], ResourceY)

    def test_subs(self):
        assert {
            ResourceB: ResourceBToResourceY,
            ResourceC: ResourceCToResourceZ,
        } == ResourceAToResourceX._subs

    def test_invalid_abstract_mapping_from_obj(self):
        # All sub_class.from_obj should be an sub_class of base_class.from_obj
        with pytest.raises(MappingSetupError) as cm:

            class _(ResourceAToResourceX):
                from_obj = ResourceY
                to_obj = ResourceZ

        assert "`from_obj` must be a subclass of `parent.from_obj`" == str(cm.value)

    def test_invalid_abstract_mapping_to_obj(self):
        # All sub_class.from_obj should be an sub_class of base_class.to_obj
        with pytest.raises(MappingSetupError) as cm:

            class _(ResourceAToResourceX):
                from_obj = ResourceC
                to_obj = ResourceB

        assert "`to_obj` must be a subclass of `parent.to_obj`" in str(cm.value)

    def test_invalid_inherited_type(self):
        with pytest.raises(TypeError) as cm:

            class ResourceD(ResourceA):
                pass

            ResourceAToResourceX.apply(ResourceD())

        assert "`source_resource` parameter must be an instance of" in str(cm.value)



"""
XML Codec (output only)

Output a resource structure as XML

XML has a unique attribute in the form of the text. This is plain text that can
be placed within a pair of tags.

To support this the XML codec includes a TextField, any TextField found on a
resource will be exported (if multiple are defined they will all be exported
into the one text block).

The TextField is for all intents and purposes just a StringField, other codecs
will export any value as a String.

"""
import datetime
from io import StringIO
from typing import TextIO
from xml.sax import saxutils

from odin import serializers, Resource
from odin import fields
from odin.fields import composite, StringField
from odin.utils import attribute_field_iter_items, element_field_iter_items, getmeta

XML_TYPES = {
    datetime.date: serializers.date_iso_format,
    datetime.time: serializers.time_iso_format,
    datetime.datetime: serializers.datetime_iso_format,
}

CONTENT_TYPE = "application/xml"


class TextField(StringField):
    """
    Special String field for representing XML text blocks.
    """


def _serialize_to_string(value):
    if value.__class__ in XML_TYPES:
        return XML_TYPES[value.__class__](value)
    else:
        return str(value)


def dump(
    fp: TextIO,
    resource,  # type: Resource
    line_ending="",  # type: str
):
    """
    Dump a resource to a file like object.
    :param fp: File pointer or file like object.
    :param resource: Resource to dump
    :param line_ending: End of line character to apply
    """
    meta = getmeta(resource)

    # Write container and any attributes
    attributes = "".join(
        " {}={}".format(
            f.name, saxutils.quoteattr(_serialize_to_string(v))
        )  # Encode attributes
        for f, v in attribute_field_iter_items(resource)
    )
    fp.write("<{}{}>{}".format(meta.name, attributes, line_ending))

    # Write any element fields
    for field, value in element_field_iter_items(resource):
        if isinstance(field, composite.ListOf):
            if field.use_container:
                fp.write("<{}>{}".format(field.name, line_ending))
            for v in value:
                dump(fp, v, line_ending)
            if field.use_container:
                fp.write("</{}>{}".format(field.name, line_ending))

        elif isinstance(field, composite.DictAs):
            if value is not None:
                dump(fp, value, line_ending)

        elif isinstance(field, fields.ArrayField):
            for v in value:
                fp.write(
                    "<{}>{}</{}>{}".format(
                        field.name, _serialize_to_string(v), field.name, line_ending
                    )
                )

        elif isinstance(field, TextField):
            if value is not None:
                fp.write(
                    "{}{}".format(
                        saxutils.escape(_serialize_to_string(value)), line_ending
                    )
                )

        else:
            fp.write(
                "<{}>{}</{}>{}".format(
                    field.name,
                    saxutils.escape(_serialize_to_string(value)),
                    field.name,
                    line_ending,
                )
            )

    fp.write("</{}>{}".format(meta.name, line_ending))


def dumps(resource, **kwargs):
    """
    Dump a resource to a string.

    :param resource: Resource to dump
    """
    f = StringIO()
    dump(f, resource, **kwargs)
    return f.getvalue()



import os
from datetime import date

from odin.codecs import xml_codec
from .resources import *

FIXTURE_PATH_ROOT = os.path.join(os.path.dirname(__file__), "fixtures")


class Summary(odin.Resource):
    format = odin.StringField(is_attribute=True)
    content = xml_codec.TextField()


class XMLBook(LibraryBook):
    title = odin.StringField()
    isbn = odin.StringField(key=True)
    num_pages = odin.IntegerField()
    rrp = odin.FloatField(default=20.4, use_default_if_not_provided=True)
    fiction = odin.BooleanField(is_attribute=True)
    genre = odin.StringField(
        choices=(
            ("sci-fi", "Science Fiction"),
            ("fantasy", "Fantasy"),
            ("biography", "Biography"),
            ("others", "Others"),
            ("computers-and-tech", "Computers & technology"),
        )
    )
    published = odin.TypedArrayField(odin.DateField())
    authors = odin.ArrayOf(Author, use_container=True)
    publisher = odin.DictAs(Publisher, null=True)
    summary = odin.DictAs(Summary)


class TestXmlLoad:
    def test_dumps(self):
        book = XMLBook(
            title="Consider Phlebas & Other stories",
            isbn="0-333-45430-8",
            num_pages=471,
            rrp=19.50,
            fiction=True,
            genre="sci-fi",
            published=[date(1987, 1, 1)],
            authors=[Author(name="Iain M. Banks")],
            publisher=Publisher(name="Macmillan"),
            summary=Summary(
                format="text/plain",
                content="The Culture and the Idiran Empire are at war in a galaxy-spanning conflict.",
            ),
        )

        assert """<XMLBook fiction="True">
<title>Consider Phlebas &amp; Other stories</title>
<isbn>0-333-45430-8</isbn>
<num_pages>471</num_pages>
<rrp>19.5</rrp>
<genre>sci-fi</genre>
<published>1987-01-01</published>
<authors>
<Author>
<name>Iain M. Banks</name>
</Author>
</authors>
<Publisher>
<name>Macmillan</name>
</Publisher>
<Summary format="text/plain">
The Culture and the Idiran Empire are at war in a galaxy-spanning conflict.
</Summary>
</XMLBook>
""" == xml_codec.dumps(
            book, line_ending="\n"
        )



import sys
from typing import Optional, Union, List, Dict, Tuple, Any, Final

import pytest

import odin
from odin.annotated_resource import type_resolution
from odin.annotated_resource.special_fields import AnyField

from ..resources_annotated import From, Book


class NonsenseField(odin.Field):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        raise TypeError("Nonsense!")

    def to_python(self, value):
        pass


class TestOptions:
    def test_kwargs__where_field_type_is_field_subclass(self):
        target = type_resolution.Options(field_type=odin.StringField, my_value=123)

        actual = target._kwargs()

        assert actual == {
            "choices": None,
            "default": odin.fields.NotProvided,
            "doc_text": "",
            "error_messages": None,
            "is_attribute": False,
            "key": False,
            "name": None,
            "use_default_if_not_provided": True,
            "validators": None,
            "verbose_name": None,
            "verbose_name_plural": None,
            "my_value": 123,
        }

    def test_kwargs__where_field_type_is_not_field_subclass(self):
        target = type_resolution.Options(field_type=odin.ConstantField, my_value=123)

        actual = target._kwargs()

        assert actual == {
            "doc_text": "",
            "is_attribute": False,
            "key": False,
            "name": None,
            "verbose_name": None,
            "verbose_name_plural": None,
            "my_value": 123,
        }

    def test_init_field__where_field_is_defined(self):
        target = type_resolution.Options(field_type=odin.StringField)

        actual = target.init_field()

        assert isinstance(actual, odin.StringField)

    def test_init_field__where_field_is_not_defined(self):
        target = type_resolution.Options()

        with pytest.raises(
            odin.exceptions.ResourceDefError,
            match="Field type `None` could not be resolved",
        ):
            target.init_field()

    @pytest.mark.skipif(
        sys.version_info < (3, 10), reason="requires python3.10 or higher"
    )
    def test_init_field__where_field_has_invalid_args(self):
        target = type_resolution.Options(
            field_type=odin.StringField, my_invalid_arg="boom!"
        )

        with pytest.raises(
            odin.exceptions.ResourceDefError,
            match=r"StringField\.__init__\(\) got an unexpected keyword argument 'my_invalid_arg'",
        ):
            target.init_field()

    @pytest.mark.skipif(
        sys.version_info < (3, 10), reason="requires python3.10 or higher"
    )
    def test_init_field__where_virtual_field_has_invalid_args(self):
        target = type_resolution.Options(
            field_type=odin.ConstantField, value="foo", my_invalid_arg="boom!"
        )

        with pytest.raises(
            odin.exceptions.ResourceDefError,
            match=r"ConstantField\.__init__\(\) got an unexpected keyword argument 'my_invalid_arg'",
        ):
            target.init_field()

    @pytest.mark.skipif(
        sys.version_info < (3, 10), reason="requires python3.10 or higher"
    )
    def test_init_field__where_field_has_missing_required_arg(self):
        target = type_resolution.Options(field_type=odin.EnumField)

        with pytest.raises(
            odin.exceptions.ResourceDefError,
            match=r"EnumField\.__init__\(\) missing 1 required positional argument: 'enum_type'",
        ):
            target.init_field()

    def test_init_field__where_unknown_type_error_raised(self):
        target = type_resolution.Options(field_type=NonsenseField)

        with pytest.raises(TypeError, match=r"Nonsense!"):
            target.init_field()


@pytest.mark.parametrize(
    "type_, expected",
    (
        (str, odin.StringField),
        (odin.types.IPv46, odin.IPv46Field),
        (From, odin.EnumField),
        (Book, odin.DictAs),
    ),
)
def test_resolve_field_from_type__where_type_can_be_resolved(type_, expected):
    options = type_resolution.Options()

    type_resolution._resolve_field_from_type(options, type_)

    assert options.field_type is expected


def test_resolve_field_from_type__where_type_cannot_be_resolved():
    options = type_resolution.Options()

    with pytest.raises(
        odin.exceptions.ResourceDefError,
        match="Unable to resolve field for type <class 'object'>",
    ):
        type_resolution._resolve_field_from_type(options, object)


@pytest.mark.parametrize(
    "type_, expected",
    (
        (Optional[str], True),
        (Union[str, None], True),
        (Union[int, float], False),
        (Union[int, float, None], False),
    ),
)
def test_is_optional(type_, expected):
    actual = type_resolution.is_optional(type_)

    assert actual is expected


@pytest.mark.parametrize(
    "type_, expected",
    (
        (Optional[str], odin.StringField),
        (Union[str, None], odin.StringField),
        (List, odin.ListField),
        (List[Book], odin.ListOf),
        (Dict, odin.DictField),
        (Dict[str, Book], odin.DictOf),
    ),
)
def test_resolve_field_from_sub_scripted_type__where_field_can_be_resolved(
    type_, expected
):
    options = type_resolution.Options()
    origin = type_resolution.get_origin(type_)

    type_resolution._resolve_field_from_sub_scripted_type(origin, options, type_)

    assert options.field_type is expected


@pytest.mark.parametrize(
    "type_, expected_field",
    (
        (
            List[str],
            odin.StringField,
        ),
        (
            List[From],
            odin.EnumField,
        ),
        (
            List[Any],
            AnyField,
        ),
    ),
)
def test_resolve_field_from_sub_scripted_type__where_field_can_be_resolved_to_typed_list(
    type_, expected_field
):
    options = type_resolution.Options()
    origin = type_resolution.get_origin(type_)

    type_resolution._resolve_field_from_sub_scripted_type(origin, options, type_)

    assert options.field_type is odin.TypedListField
    assert isinstance(options.field_args["field"], expected_field)


@pytest.mark.parametrize(
    "type_, expected_key_field, expected_value_field",
    (
        (
            Dict[str, str],
            odin.StringField,
            odin.StringField,
        ),
        (
            Dict[str, From],
            odin.StringField,
            odin.EnumField,
        ),
        (
            Dict[str, Any],
            odin.StringField,
            AnyField,
        ),
    ),
)
def test_resolve_field_from_sub_scripted_type__where_field_can_be_resolved_to_typed_dict(
    type_, expected_key_field, expected_value_field
):
    options = type_resolution.Options()
    origin = type_resolution.get_origin(type_)

    type_resolution._resolve_field_from_sub_scripted_type(origin, options, type_)

    assert options.field_type is odin.TypedDictField
    assert isinstance(options.field_args["key_field"], expected_key_field)
    assert isinstance(options.field_args["value_field"], expected_value_field)


@pytest.mark.parametrize(
    "type_",
    (
        Union[int, float],
        Tuple[int, float],
        Tuple,
    ),
)
def test_resolve_field_from_sub_scripted_type__where_type_is_unsupported(type_):
    options = type_resolution.Options()
    origin = type_resolution.get_origin(type_)

    with pytest.raises(
        odin.exceptions.ResourceDefError,
        match="Unable to resolve field for sub-scripted type",
    ):
        type_resolution._resolve_field_from_sub_scripted_type(origin, options, type_)


@pytest.mark.parametrize(
    "type_, expected_type, expected_args",
    (
        # Basic types
        (str, odin.StringField, {}),
        (list, odin.ListField, {}),
        (dict, odin.DictField, {}),
        (odin.types.Email, odin.EmailField, {}),
        (Optional[odin.types.Url], odin.UrlField, {"null": True}),
        # List/Dict types
        (List[Book], odin.ListOf, {"resource": Book}),
        (Dict[str, Book], odin.DictOf, {"resource": Book}),
        # Enums
        (From, odin.EnumField, {"enum_type": From}),
        (Optional[From], odin.EnumField, {"enum_type": From, "null": True}),
    ),
)
def test_resolve_field_from_annotation__where_field_can_be_resolved(
    type_, expected_type, expected_args
):
    options = type_resolution.Options()

    type_resolution._resolve_field_from_annotation(options, type_)
    actual_args = options._kwargs()

    assert options.field_type is expected_type
    assert all(
        actual_args.get(key) == value for key, value in expected_args.items()
    ), f"Expected args {expected_args} not found in {actual_args}"


def test_resolve_field_from_annotation__for_list_field():
    options = type_resolution.Options()

    type_resolution._resolve_field_from_annotation(options, List[bool])
    actual_args = options._kwargs()

    assert options.field_type is odin.TypedListField
    assert actual_args["default"] is list
    assert isinstance(actual_args["field"], odin.BooleanField)


def test_resolve_field_from_annotation__for_dict_field():
    options = type_resolution.Options()

    type_resolution._resolve_field_from_annotation(options, Dict[str, int])
    actual_args = options._kwargs()

    assert options.field_type is odin.TypedDictField
    assert actual_args["default"] is dict
    assert isinstance(actual_args["key_field"], odin.StringField)
    assert isinstance(actual_args["value_field"], odin.IntegerField)


def test_resolve_field_from_annotation__for_nested_dict_types():
    options = type_resolution.Options()

    type_resolution._resolve_field_from_annotation(options, Dict[str, List[int]])
    actual_args = options._kwargs()

    assert options.field_type is odin.TypedDictField
    assert actual_args["default"] is dict
    assert isinstance(actual_args["key_field"], odin.StringField)
    assert isinstance(actual_args["value_field"], odin.TypedListField)
    # Nested field type
    assert isinstance(actual_args["value_field"].field, odin.IntegerField)


def test_resolve_field_from_annotation__not_a_type():
    options = type_resolution.Options()

    with pytest.raises(
        odin.exceptions.ResourceDefError, match="Annotation is not a type instance"
    ):
        type_resolution._resolve_field_from_annotation(options, object())


@pytest.mark.parametrize(
    "type_, value, expected_type, expected_default",
    (
        (str, odin.NotProvided, odin.StringField, odin.NotProvided),
        (str, "foo", odin.StringField, "foo"),
        # Defined as a field
        (str, odin.IntegerField(default=123), odin.IntegerField, 123),
        # Defined with Options
        (str, odin.Options("bar"), odin.StringField, "bar"),
        # Explicit type provided to options
        (
            str,
            odin.Options(field_type=odin.IntegerField),
            odin.IntegerField,
            odin.NotProvided,
        ),
    ),
)
def test_process_attribute__where_assumed_valid(
    type_, value, expected_type, expected_default
):
    actual = type_resolution.process_attribute(type_, value)

    assert isinstance(actual, expected_type)
    assert actual.default == expected_default


class TestSpecialFields:
    @pytest.mark.parametrize(
        "target, value, expected",
        (
            (type_resolution.AnyField(), None, None),
            (type_resolution.AnyField(), "abc", "abc"),
            (type_resolution.AnyField(), 123, 123),
        ),
    )
    def test_to_python(self, target, value, expected):
        actual = target.to_python(value)

        assert actual == expected

    def test_constant_field__where_field_is_defined_correctly(self):
        actual = type_resolution.process_attribute(Final[str], "foo")

        assert isinstance(actual, odin.ConstantField)
        assert actual.value == "foo"

    def test_constant_field__where_value_is_missing(self):
        with pytest.raises(
            odin.exceptions.ResourceDefError, match="Final fields require a value"
        ):
            type_resolution.process_attribute(Final[str], odin.NotProvided)



"""
Annotated Resources
~~~~~~~~~~~~~~~~~~~

Annotated resources make use of Python type annotations to simplify the
definition of data structures.

Annotated resources still utilise many of the existing classes eg
``ResourceBase``, ``ResourceOptions`` and all existing fields. Allowing
existing code (eg Mappings, Codecs) including custom implementations to be
utilised unchanged.

"""
import copy
from typing import Any, Dict, Tuple, Type, Optional, TypeVar, Iterable

from odin import registration
from odin.fields import BaseField
from odin.resources import (
    ResourceBase,
    ResourceOptions,
    DEFAULT_TYPE_FIELD,
    NotProvided,
)
from .type_resolution import process_attribute, Options
from ..exceptions import ResourceDefError

__all__ = (
    "Options",
    "AnnotatedResourceType",
    "AnnotatedResource",
    "AResource",
)

from ..utils import getmeta

MOT = TypeVar("MOT", bound=ResourceOptions)


def _new_meta_instance(
    meta_options_type: Type[MOT],
    meta_def: Optional[object],
    new_class: "AnnotatedResourceType",
) -> MOT:
    """
    Instantiate meta options instance and handle inheritance of required fields
    """
    base_meta = getattr(new_class, "_meta", None)
    new_meta = meta_options_type(meta_def)
    new_class.add_to_class("_meta", new_meta)

    # Namespace is inherited
    if base_meta and new_meta.name_space is NotProvided:
        new_meta.name_space = base_meta.name_space
    if new_meta.name_space is NotProvided:
        new_meta.name_space = new_class.__module__

    # Type field is inherited and default if not provided
    if base_meta and new_meta.type_field is NotProvided:
        new_meta.type_field = base_meta.type_field
    if new_meta.type_field is NotProvided:
        new_meta.type_field = DEFAULT_TYPE_FIELD

    # Key field names is inherited
    if base_meta and new_meta.key_field_names is None:
        new_meta.key_field_names = base_meta.key_field_names

    # Field sorting is inherited
    if new_meta.field_sorting is NotProvided:
        new_meta.field_sorting = base_meta.field_sorting if base_meta else False

    return new_meta


def _iterate_attrs(attrs: Dict[str, Any]) -> Iterable[Tuple[str, BaseField]]:
    """
    Iterate through attributes and combine with annotations
    """
    annotations = attrs.pop("__annotations__", None) or {}

    # Yield any annotations processed into field instances
    for name, type_ in annotations.items():
        # Ignore all uppercase entries (consts)
        if not name.isupper():
            value = attrs.pop(name, NotProvided)
            yield name, process_attribute(type_, value)

    # Process any leftover fields
    yield from attrs.items()


def _add_parent_fields_to_class(
    new_class: "AnnotatedResourceType", new_meta: ResourceOptions, parents
):
    """
    Iterate through parent attrs and yield fields
    """
    # All the fields of any type declared on this model
    local_field_attr_names = {f.attname for f in new_meta.fields}
    field_attr_names = set(local_field_attr_names)

    for base, base_meta in (
        (base, getmeta(base)) for base in parents if hasattr(base, "_meta")
    ):
        # Check for clashes between locally declared fields and those
        # on the base classes (we cannot handle shadowed fields at the
        # moment).
        for field in base_meta.all_fields:
            if field.attname in local_field_attr_names:
                raise ResourceDefError(
                    f"Local field {field.attname!r} in class {new_class.__name__!r} "
                    f"clashes with field of the same name from base class {base.__name__!r}"
                )

        # Clone fields (but filter out fields already inherited)
        for field in (
            field for field in base_meta.fields if field.attname not in field_attr_names
        ):
            field_attr_names.add(field.attname)
            new_class.add_to_class(field.attname, copy.deepcopy(field))

        # Clone any virtual fields
        for field in base_meta.virtual_fields:
            new_class.add_to_class(field.attname, copy.deepcopy(field))

        # Add to parents list
        new_meta.parents += base_meta.parents
        new_meta.parents.append(base)


class AnnotatedResourceType(type):
    def __new__(
        mcs,
        name: str,
        bases,
        attrs: dict,
        meta_options_type: Type[MOT] = ResourceOptions,
        abstract: bool = False,
    ):
        super_new = super().__new__

        # attrs will never be empty for classes declared in the standard way
        # (i.e. with the `class` keyword). This is quite robust.
        if name == "NewBase" and attrs == {}:
            return super_new(mcs, name, bases, attrs)

        parents = [
            base
            for base in bases
            if (
                isinstance(base, AnnotatedResourceType)
                and not (base.__name__ == "NewBase" and base.__mro__ == (base, object))
            )
        ]
        if not parents:
            # If this isn't a subclass of NewResource, don't do anything special.
            return super_new(mcs, name, bases, attrs)

        # Create the class.
        new_class = super_new(mcs, name, bases, {"__module__": attrs.pop("__module__")})

        # Create new meta instance
        new_meta = _new_meta_instance(
            meta_options_type, attrs.pop("Meta", None), new_class
        )

        # Bail out early if we have already created this class.
        r = registration.get_resource(new_meta.resource_name)
        if r is not None:
            return r

        # Add all field attributes to the class.
        for name, field in _iterate_attrs(attrs):
            new_class.add_to_class(name, field)

        _add_parent_fields_to_class(new_class, new_meta, parents)

        # Sort the fields
        if not new_meta.field_sorting:
            new_meta.fields = sorted(new_meta.fields, key=hash)

        # Give fields an opportunity to do additional operations after the
        # resource is full populated and ready.
        for field in new_meta.all_fields:
            if hasattr(field, "on_resource_ready"):
                field.on_resource_ready()

        if abstract:
            return new_class

        # Register resource
        registration.register_resources(new_class)

        # Because of the way imports happen (recursively), we may or may not be
        # the first time this model tries to register with the framework. There
        # should only be one class for each model, so we always return the
        # registered version.
        return registration.get_resource(new_meta.resource_name)

    def add_to_class(cls, name, value):
        if hasattr(value, "contribute_to_class"):
            value.contribute_to_class(cls, name)
        else:
            setattr(cls, name, value)


class AnnotatedResource(
    ResourceBase, metaclass=AnnotatedResourceType, meta_options_type=ResourceOptions
):
    """
    New Style Resource utilising type annotations for defining fields
    """


AResource = AnnotatedResource



def sum_fields(*field_values):
    """
    Return a sum of fields

    :param field_values:
    :return:

    """
    return sum(field_values)


class JoinFields:
    """
    Helper for combining multiple fields
    """

    __slots__ = ("sep",)

    def __init__(self, sep=""):
        self.sep = sep

    def __call__(self, *field_values):
        return self.sep.join(field_values)


join_fields = JoinFields
# Backwards compatibility
cat_fields = CatFields = JoinFields


class SplitField:
    """
    Helper for splitting a field into multiple fields.
    """

    __slots__ = (
        "sep",
        "max_split",
    )

    def __init__(self, sep: str = None, max_split: int = None):
        self.sep = sep
        self.max_split = max_split

    def __call__(self, field_value):
        if self.max_split is None:
            return field_value.split(self.sep)
        else:
            return field_value.split(self.sep, self.max_split)


split_field = SplitField


class ApplyMapping:
    """
    Helper for applying a mapper.

    This helper should be used along with the bind flag so the context object can be maintained.
    """

    def __init__(self, mapping, allow_subclass=False):
        self.mapping = mapping
        self.allow_subclass = allow_subclass

    def __call__(self, bound_self, field_value):
        # Note this returns a tuple, this is expected
        return (
            self.mapping.apply(
                field_value,
                context=bound_self.context,
                allow_subclass=self.allow_subclass,
            ),
        )

    def __repr__(self):
        return "<{}: {}.{}>".format(
            self.__class__.__name__, self.mapping.__module__, self.mapping.__name__
        )


MapDictAs = MapListOf = ApplyMapping


class NoOpMapper:
    """
    Helper that provides the mapper interface performs no operation on the object.

    This is used with the MapListOf and MapDictAs fields when both contain the same Resource type.
    """

    @staticmethod
    def apply(source_obj, **_):
        return source_obj



import math

__all__ = ("latitude", "longitude", "latlng", "point")

from typing import Tuple


def to_dms(value, absolute=False):
    # type: (float, bool) -> Tuple[float, float, float]
    """
    Split a float value into DMS (degree, minute, second) parts

    :param value: Float value to split
    :param absolute: Obtain the absolute value
    :return: tuple containing DMS values

    """
    invert = not absolute and value < 0
    value = abs(value)
    degree = int(math.floor(value))
    value = (value - degree) * 60
    minute = int(math.floor(value))
    second = (value - minute) * 60
    return (-degree if invert else degree), minute, second


def to_dm(value, absolute=False):
    # type: (float, bool) -> Tuple[float, float]
    """
    Split a float value into DM (degree, minute) parts

    :param value: Float value to split
    :param absolute: Obtain the absolute value
    :return: tuple containing DM values

    """
    invert = not absolute and value < 0
    value = abs(value)
    degree = int(math.floor(value))
    minute = (value - degree) * 60
    return (-degree if invert else degree), minute


class latitude(float):  # NoQA
    """
    A latitude value. A latitude is a value between -90.0 and 90.0.
    """

    def __new__(cls, x):
        lat = float.__new__(cls, x)
        if lat > 90.0 or lat < -90.0:
            raise ValueError(f"not in range -90.0 -> 90.0: '{x}'")
        return lat

    def __str__(self):
        result = "{:02d}°{:02d}'{:02f}\"".format(*to_dms(self, True))
        return result + ("S" if self < 0 else "N")


class longitude(float):  # NoQA
    """
    A longitude value. A longitude is a value between -180.0 and 180.0.
    """

    def __new__(cls, x):
        lng = float.__new__(cls, x)
        if lng > 180.0 or lng < -180.0:
            raise ValueError("not in range -180.0 -> 180.0: '{}'".format(x))
        return lng

    def __str__(self):
        result = "{:03d}°{:02d}'{:02f}\"".format(*to_dms(self, True))
        return result + ("W" if self < 0 else "E")


class latlng(tuple):
    """
    Combination latitude and longitude value.
    """

    def __new__(cls, *args):
        # Unpack a tuple, list or latlng instance.
        if len(args) == 1 and isinstance(args[0], (tuple, list)):
            args = args[0]
        try:
            lat, lng = args
        except ValueError:
            raise TypeError(f"latlng() takes 2 arguments ({len(args)} given)")
        return tuple.__new__(cls, (latitude(lat), longitude(lng)))

    @property
    def lat(self):
        return self[0]

    @property
    def lng(self):
        return self[1]

    def __repr__(self):
        return "latlng({!r}, {!r})".format(*self)

    def __str__(self):
        return "({}, {})".format(*self)


class point(tuple):
    """
    A point in cartesian space. This type can be either 2D (on a plain) or 3D (includes a z-axis).
    """

    def __new__(cls, *args):
        if len(args) == 1 and isinstance(args[0], (tuple, list)):
            args = args[0]
        if len(args) in (2, 3):
            return tuple.__new__(cls, (float(x) for x in args))
        else:
            raise TypeError(f"point() takes 2-3 arguments ({len(args)} given)")

    @property
    def x(self):
        return self[0]

    @property
    def y(self):
        return self[1]

    @property
    def z(self):
        try:
            return self[2]
        except IndexError:
            raise AttributeError("2D points do not include a z axis.")

    @property
    def is_3d(self):
        return len(self) == 3

    def __repr__(self):
        return ("point({}, {}, {})" if self.is_3d else "point({}, {})").format(*self)

    def __str__(self):
        return ("({}, {}, {})" if self.is_3d else "({}, {})").format(*self)
